#include "../include/config.hpp"
#include "../include/blender.hpp"
#include "../include/vector_parser.hpp"
#include "../include/return_heap_parser.hpp"
#include "../include/ObjectParser.hpp"

#include <boost/python/detail/api_placeholder.hpp>

namespace MQO {

using phoenix::bind;
using phoenix::arg1;
using phoenix::static_cast_;

double scale_ = 0.1;

static py::object callMultMatrix( py::object x, py::object y )
{
	PyObject *px = x.ptr();
	PyObject *py = y.ptr();
	py::object z( py::handle<> (px->ob_type->tp_as_number->nb_multiply( px, py )) );
	return z;
}

struct Assigner
{
	typedef void result_type;
	typedef boost::shared_ptr<void> pointer_t;
	typedef boost::array<double,3> vector_t;

	void operator()( py::object nmesh, py::object bobject,
				ObjectParser::hierarchy_t const *hierarchy, int key, pointer_t const &val, PropValue *propVal )
	{
		switch( key ){
		case ObjectParser::Depth:
			{
				int depth = *reinterpret_cast<unsigned int*>( val.get() );
				const_cast<ObjectParser::hierarchy_t*>(hierarchy)->push_back(
					std::make_pair(depth, bobject) );
				return;
			}
		case ObjectParser::Scale:
			{
				vector_t &sz = *reinterpret_cast<vector_t*>( val.get() );
				py::list col0 = py::list( py::make_tuple( sz[0], 0.0, 0.0 ) );
				py::list col1 = py::list( py::make_tuple( 0.0, sz[2], 0.0 ) );
				py::list col2 = py::list( py::make_tuple( 0.0, 0.0, sz[1] ) );
				py::object scmat = Blender.Mathutils().attr("Matrix")( col0, col1, col2 );
				bobject.attr("setMatrix")( callMultMatrix(
					bobject.attr("matrix"), scmat.attr("resize4x4")() ) );
				return;
			}

		case ObjectParser::Rotation:
			{
				vector_t &rot = *reinterpret_cast<vector_t*>( val.get() );
				py::list ang = py::list( py::make_tuple( rot[0], rot[2], rot[1] ) );

				py::object euler = Blender.Mathutils().attr("Euler")( ang );
				py::object rotmat = euler.attr("toMatrix")().attr("resize4x4")();
				bobject.attr("setMatrix")( callMultMatrix( bobject.attr("matrix"), rotmat ) );
				return;
			}
		case ObjectParser::Translation:
			{
				vector_t &rot = *reinterpret_cast<vector_t*>( val.get() );
				py::list pos = py::list( py::make_tuple( 
					rot[0] * scale_, rot[2] * -scale_, rot[1] * scale_ ) );

				py::object vec = Blender.Mathutils().attr("Vector")( pos );
				py::object transmat = Blender.Mathutils().attr("TranslationMatrix")( vec );
				bobject.attr("setMatrix")( callMultMatrix( bobject.attr("matrix"), transmat ) );
			}
		case ObjectParser::Segment:
			propVal->segment = *(unsigned*)val.get();
			return;

		case ObjectParser::Visible:
			propVal->visible = *(unsigned*)val.get();
			return;

		case ObjectParser::Patch:
			propVal->patch = *(unsigned*)val.get();
			return;

		case ObjectParser::Mirror:
			propVal->mirror = *(unsigned*)val.get();
			return;

		case ObjectParser::MirrorAxis:
			propVal->mirror_axis = *(unsigned*)val.get();
			return;

		case ObjectParser::MirrorDis:
			propVal->mirror_dis = *(double*)val.get();
			return;

		case ObjectParser::Shading:
			propVal->shading = *(unsigned*)val.get();
			return;

		case ObjectParser::Facet:
			propVal->facet = (unsigned)*(double*)val.get();
			return;
		}
	}
};

struct PropertyParser
{
	typedef boost::shared_ptr<void> pointer_t;

	typedef stored_rule< scanner_t, parser_context<pointer_t> > property_rule;
	struct property_closure:closure<
		property_closure, int, typename property_rule::alias_t>
	{
		member1 key;
		member2 target_r;
	};
	typedef std::pair<int,property_rule> kr;

	symbols< kr > selector_r;

	PropertyParser()
	{
		selector_r.add
		("depth",		kr( ObjectParser::Depth,		heap_p(uint_p)))
		("folding",		kr( ObjectParser::Folding,		heap_p(uint_p)))
		("patch",		kr( ObjectParser::Patch,		heap_p(uint_p)))
		("segment",		kr( ObjectParser::Segment,		heap_p(uint_p)))
		("visible",		kr( ObjectParser::Visible,		heap_p(uint_p)))
		("locking",		kr( ObjectParser::Locking,		heap_p(uint_p)))
		("shading",		kr( ObjectParser::Shading,		heap_p(uint_p)))
		("facet",		kr( ObjectParser::Facet,		heap_p(real_p)))
		("color",		kr( ObjectParser::Color,		heap_p( vector_p<3>( real_p) )))
		("color_type",	kr( ObjectParser::ColorType,	heap_p(uint_p)))
		("mirror",		kr( ObjectParser::Mirror,		heap_p(uint_p)))
		("mirror_axis",	kr( ObjectParser::MirrorAxis,	heap_p(uint_p)))
		("mirror_dis",	kr( ObjectParser::MirrorDis,	heap_p(real_p)))
		("lathe",		kr( ObjectParser::Lathe,		heap_p(uint_p)))
		("lathe_axis",	kr( ObjectParser::LatheAxis,	heap_p(uint_p)))
		("lathe_seg",	kr( ObjectParser::LatheSeg,		heap_p(uint_p)))
		("scale",		kr( ObjectParser::Scale,		heap_p( vector_p<3>( real_p) )))
		("rotation",	kr( ObjectParser::Rotation,		heap_p( vector_p<3>( real_p) )))
		("translation",	kr( ObjectParser::Translation,	heap_p( vector_p<3>( real_p) )))
		;
	}

	parser_result<rule<scanner_t, property_closure::context_t>,scanner_t>::type
	parse( scanner_t const &scan, py::object nmesh, py::object bobject,
										ObjectParser::hierarchy_t const *hierarchy, PropValue *propVal )
	{
		rule<scanner_t, property_closure::context_t> property_p;
		property_p = *(
			selector_r
			[
				property_p.key =
					bind(&kr::first)(static_cast_<kr&>(arg1)),
				property_p.target_r =
					bind(&kr::second)(static_cast_<kr&>(arg1))
			]
			>> lazy_p(property_p.target_r)
			[
				bind(Assigner())(nmesh, bobject, hierarchy, property_p.key, arg1, propVal)
			]
		);
		return property_p.parse(scan);
	}
};



bool ObjectParser::parseProperty(
	scanner_t const &scan,
	py::object nmesh,
	py::object bobject,
	ObjectParser::hierarchy_t const *hierarchy,
	PropValue *propVal
	 )const
{
	static PropertyParser property_p;
	scale_ = scale;

	memset(propVal, 0, sizeof(*propVal));
	propVal->visible = 1;

	return property_p.parse( scan, nmesh, bobject, hierarchy, propVal ).operator_bool();
}

}//MQO